/* 
 * MySQL 4.1.x NEB module for nagios 2.0 by Matthew Kent based on work by Ben
 * 
 *
 * Modified by Billy Gunn (billy@gunn.org) for use with
 * Nagios Plugin for Cacti.
 * 	
 *	Changes:
 *		- Table structure
 *		- Modified to work with Nagios 2.0b4
 *		- Added support for servicegroups
 *
 *
 *
 *
 * Notes about this implementation.
 *
 * It has some attempts at speeding things up with the bulking of inserts.
 * The time it will wait between database runs while buffering queries is tunable
 * but not yet support by passing in an argument from nagios.cfg proper.
 * 
 * If you're looking for something to customize yourself, I would definately 
 * recommend the postgres implementation as this one has obvious shortcomings 
 * in abstracting the code from the database logic. Until mysql 5 is production
 * quality, this will remain pretty ugly. 
 *
 * TODO:
 *  - Passing arguments from nagios for certain options.
 *  - Adding a verbose mode.
 *  - Cleanup sql statements for uniformity.
 *  - Error in sql before loading startup host/service data causes multiple
 *    'inserter unloading' msg's
 *  - General tidying.
 */
#ifndef NSCORE
#define NSCORE
#endif

#define _GNU_SOURCE

/* we need this for mysql support */
#include <mysql/mysql.h>
#include <mysql/errmsg.h>

/* include the needed event broker header files */
#include "../include/nebmodules.h"
#include "../include/nebcallbacks.h"
#include "../include/nebstructs.h"
#include "../include/broker.h"

/* include some Nagios stuff as well */
#include "../include/config.h"
#include "../include/common.h"
#include "../include/nagios.h"
#include "../include/objects.h"

#include <sys/types.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>

/* specify event broker API version (required) */
NEB_API_VERSION(CURRENT_NEB_API_VERSION);

#define NULL_STRING(x) (x?:"null")
#define BOOLIFY_STRING(x) (x?"TRUE":"FALSE")

#define ERR_QUERY_BUILD "Building query failed, aborting"

typedef struct{
	char host[512];
	char user[512];
	char pass[512];
	char database[512];
} DBCONNECTION;

char *querify(char * src);
int append(char **, char*, int);
int buffer(char ***, char*, int);
void err_out(char *);
void free_buffers();
void *inserter_module_handle=NULL;

extern hostgroup *hostgroup_list;
extern servicegroup *servicegroup_list;
extern host *host_list;
extern service *service_list;

extern int enable_flap_detection;
extern int enable_notifications;
extern int enable_event_handlers;
extern int execute_service_checks;
extern int execute_host_checks;
extern int accept_passive_service_checks;
extern int accept_passive_host_checks;

static int loadconfig(int, void *);
static int processStatus(int, void *);
static int processCheck(int, void *);
static int sql_dump();

MYSQL mysql;

#define BULK_TOTAL 4

#define BULK_HOST 0
#define BULK_HOST_CHECK 1
#define BULK_SERVICE 2
#define BULK_SERVICE_CHECK 3

/* persistant buffer for our bulkable inserts 
 * these are dependant on order, so we process
 * them in the same sequence on each run */
char *bulk_buf[BULK_TOTAL];
size_t bulk_buf_count = BULK_TOTAL;

/* short-term query buffer for our non-bulkable queries, 
 * these are not dependant on ordering like the host/service data */
char **query_buf = NULL;
size_t query_buf_count = 0;

/* how often to schedule a database run */
// todo: fill from arguments
int interval = 1;
int debug = 1;

int delayed_unload = FALSE;
/**
 * Fix 18.07.2006 / dp1
 *
 * Read configuration from file /etc/nagios2/npc.cfg" instead of hard-coding it...
 */
int nebmodule_getdbconfig(DBCONNECTION *dbconn, char *fname){
	int 	buffsize = 512;
	char	*conf_separator = "=";
	char  	*token;
	char 	buffer[buffsize];
	char 	error[2048];
	FILE 	*fp = fopen(fname, "r");
	if (fp == NULL){
		sprintf(error,"nagios-db: inserter failed to open config file (%s)", fname);
		write_to_logs_and_console(error,NSLOG_INFO_MESSAGE,TRUE);
		return 0;
	}
	while (fgets(buffer, buffsize, fp)){
		token = strtok(buffer, conf_separator);
		if (strcasecmp(token,"host") == 0){
			token = strtok(NULL, conf_separator);
			token = strtok(token, "\n");
			strcpy(dbconn->host,token);
			continue;
		}
		if (strcasecmp(token,"user") == 0){
			token = strtok(NULL, conf_separator);
			token = strtok(token, "\n");
			strcpy(dbconn->user,token);
			continue;
		}
		if (strcasecmp(token,"pass") == 0){
			token = strtok(NULL, conf_separator);
			token = strtok(token, "\n");
			strcpy(dbconn->pass,token);
			continue;
		}
		if (strcasecmp(token,"database") == 0){
			token = strtok(NULL, conf_separator);
			token = strtok(token, "\n");
			strcpy(dbconn->database,token);
			continue;
		}
	}
	return 1;
}

/* this function gets called when the module is loaded by the event broker */
int nebmodule_init(int flags, char *args, nebmodule *handle)
{
	char error[2048];
	char *fname = "/etc/nagios2/npc.cfg";
	time_t current_time;
	/**
	 * Fix 18.07.2006 / dp1
	 */
	DBCONNECTION dbconn;
	if (args != NULL){
		fname = (char *) malloc(sizeof(char) * strlen(args));
		fname = strtok(args, " ");
	}

    /* save our handle */
    inserter_module_handle=handle;

	write_to_logs_and_console("initializing nagios-db mysql inserter...",NSLOG_INFO_MESSAGE,TRUE);
	
	time(&current_time);

	/* call dump every least once every interval seconds */
	schedule_new_event(EVENT_USER_FUNCTION,TRUE,current_time+interval,TRUE,interval,NULL,TRUE,sql_dump,NULL);
	
	mysql_init(&mysql);
	
	/**
	 * Fix 18.07.2006 / dp1
	 */
	if (!nebmodule_getdbconfig(&dbconn, fname)){
		write_to_logs_and_console("nagios-db: initialization aborted due to previous errors",NSLOG_INFO_MESSAGE,TRUE);
		exit(1);
	}
	if (!mysql_real_connect(&mysql,dbconn.host,dbconn.user,dbconn.pass,dbconn.database,3306,NULL,CLIENT_MULTI_STATEMENTS)) {	
	/**
	 * End-Fix
	 */
		sprintf(error,"nagios-db: inserter failed to connect to db (%s), disabling module!", mysql_error(&mysql));
		write_to_logs_and_console(error,NSLOG_INFO_MESSAGE,TRUE);
	} else {
	
	write_to_logs_and_console("nagios-db: inserter locked into db",NSLOG_INFO_MESSAGE,TRUE);

	/* do startup routine after the point when nagios loads its retention file */
	neb_register_callback(NEBCALLBACK_TIMED_EVENT_DATA, inserter_module_handle, 0, loadconfig);
	}

	return 0;
}


/* this function gets called when the module is unloaded by the event broker */
int nebmodule_deinit(int flags, int reason)
{

	deschedule_event(EVENT_USER_FUNCTION, TRUE, sql_dump, NULL);

	write_to_logs_and_console("nagios-db: inserter unloading...",NSLOG_INFO_MESSAGE,TRUE);

	mysql_close(&mysql);

	neb_deregister_callback(NEBCALLBACK_HOST_STATUS_DATA, processStatus);
	neb_deregister_callback(NEBCALLBACK_SERVICE_STATUS_DATA, processStatus);
        neb_deregister_callback(NEBCALLBACK_HOST_CHECK_DATA, processCheck);
	neb_deregister_callback(NEBCALLBACK_SERVICE_CHECK_DATA, processCheck);

	return 0;
}

/* when Nagios starts up, we'll want to replicate configuration data into the database */
static int loadconfig(int cmd, void * data)
{
	host		*hl = 0;
	service		*sl = 0;
	hostgroup	*hg = 0;
	servicegroup	*sg = 0;
	int		groupcount = 0;
	char		log[2048];
	char *tmp;

	/* verify that we're dealing with the right kind of message. */
	if(cmd != NEBCALLBACK_TIMED_EVENT_DATA) return 0;

	/* verify that our config data has been initialized. */
	if(((nebstruct_timed_event_data*)data)->type != NEBTYPE_TIMEDEVENT_ADD) return 0;

	/* Now that we have a database connection, clear out the previous config info */

	/* clear out our memberships, we reload these each time */
	asprintf(&tmp,
		"delete from npc_hostgroup_membership");
		// err handling
	buffer(&query_buf, tmp, ++query_buf_count);
	free(tmp);

	/* clear out non persistent comments */
	asprintf(&tmp,
		"delete from npc_comment where persist = 0");
		// err handling
	buffer(&query_buf, tmp, ++query_buf_count);
	free(tmp);

	asprintf(&tmp,
		"delete from npc_servicegroup_membership");
		// err handling
	buffer(&query_buf, tmp, ++query_buf_count);
	free(tmp);

	asprintf(&tmp,
		"delete from npc_config_params");
		// err handling
	buffer(&query_buf, tmp, ++query_buf_count);
	free(tmp);

	/* Clear out npc_service_check. */
	asprintf(&tmp,
		"delete from npc_service_check");
		// err handling
	buffer(&query_buf, tmp, ++query_buf_count);
	free(tmp);

	/* this will leave any dead (removed from config hosts/hostgroups/servicegroups) as configured = false */
	asprintf(&tmp, "update npc_host,npc_service set npc_host.configured = false, npc_service.configured = false");
	// err handling
	buffer(&query_buf, tmp, ++query_buf_count);
	free(tmp);

	asprintf(&tmp, "update npc_hostgroup set npc_hostgroup.configured = false");
	// err handling
	buffer(&query_buf, tmp, ++query_buf_count);
	free(tmp);

	asprintf(&tmp, "update npc_servicegroup set npc_servicegroup.configured = false");
	// err handling
	buffer(&query_buf, tmp, ++query_buf_count);
	free(tmp);
	


	/* run out these queries right now, we need to mark all host/services as unconfigured
	 * before the large initial update which sets them back to TRUE as to weed out any dead
	 * ones to delete at the next restart */	
	sql_dump();
	
	/* update database against retained nagios data (mark configured hosts and services as such) */

	for(groupcount=0, hl = host_list; hl; hl = hl->next)
	{
		/* update this host */
		nebstruct_host_status_data ds;
		ds.object_ptr=(void *)hl;
		
		// todo: could see if this host is already in the db, if yes than set configured = true 
		// and assume the rest of the data is correct else pass to processStatus
		processStatus(NEBCALLBACK_HOST_STATUS_DATA, (void *)&ds);
		
		groupcount++;
	}

	sprintf(log,"nagios-db: configured %d hosts", groupcount);
	write_to_logs_and_console(log,NSLOG_INFO_MESSAGE,TRUE);

	for(groupcount=0, sl = service_list; sl; sl = sl->next)
	{
		/* update this service */
		nebstruct_service_status_data ds;
		ds.object_ptr=(void *)sl;
		
		processStatus(NEBCALLBACK_SERVICE_STATUS_DATA, (void *)&ds);
		
		groupcount++;
	}

	sprintf(log,"nagios-db: configured %d services", groupcount);
	write_to_logs_and_console(log,NSLOG_INFO_MESSAGE,TRUE);

	/* now that we've loaded (possibly from the retention file) flag that they haven't been checked */
	asprintf(&tmp, "update npc_host,npc_service set npc_host.has_been_checked = false, npc_service.has_been_checked = false");
	// err handling
	buffer(&query_buf, tmp, ++query_buf_count);
	free(tmp);

	/* .... and insert new config info. Start with hostgroups. */
	for(groupcount=0, hg = hostgroup_list; hg ; hg = hg->next)
	{
		hostgroupmember *hgm = 0;
		char *groupName, *groupAlias, *hostName;

		/* insert this hostgroup */
		groupName = querify(hg->group_name);
		groupAlias = querify(hg->alias);

		asprintf(&tmp,
			"INSERT INTO npc_hostgroup ("
				"name,"
				"alias,"
				"configured"
			") VALUES (%s,%s,TRUE)"
			" ON DUPLICATE KEY UPDATE "
				"alias=VALUES(alias),"
				"configured=TRUE",
				NULL_STRING(groupName),
				NULL_STRING(groupAlias));
		// err handling
		buffer(&query_buf, tmp, ++query_buf_count);
		free(tmp);
			
		if(groupAlias) free(groupAlias);

		for(hgm = hg->members; hgm; hgm = hgm->next)
		{
			/* add this host to the hostgroup */
			hostName = querify(hgm->host_name);

			asprintf(&tmp,
				"INSERT INTO npc_hostgroup_membership ("
					"hostid,"
					"groupid"
				") VALUES ((SELECT id from npc_host where name=%s),(SELECT id from npc_hostgroup where name=%s))",
					NULL_STRING(hostName),
					NULL_STRING(groupName));
			// err handling
			buffer(&query_buf, tmp, ++query_buf_count);
			free(tmp);
			
			if(hostName) free(hostName);
		}

		if(groupName) free(groupName);

		groupcount++;
	}

	sprintf(log,"nagios-db: loaded %d hostgroups", groupcount);
	write_to_logs_and_console(log,NSLOG_INFO_MESSAGE,TRUE);


	// Servicegroups
	for(groupcount=0, sg = servicegroup_list; sg ; sg = sg->next)
	{
		servicegroupmember *sgm = 0;
		char *groupName, *groupAlias, *serviceName, *hostName;

		/* insert this servicegroup */
		groupName = querify(sg->group_name);
		groupAlias = querify(sg->alias);

		asprintf(&tmp,
			"INSERT INTO npc_servicegroup ("
				"name,"
				"alias,"
				"configured"
			") VALUES (%s,%s,TRUE)"
			" ON DUPLICATE KEY UPDATE "
				"alias=VALUES(alias),"
				"configured=TRUE",
				NULL_STRING(groupName),
				NULL_STRING(groupAlias));
		// err handling
		buffer(&query_buf, tmp, ++query_buf_count);
		free(tmp);
			
		if(groupAlias) free(groupAlias);

		for(sgm = sg->members; sgm; sgm = sgm->next)
		{
			/* add this service to the servicegroup */
			hostName = querify(sgm->host_name);
			serviceName = querify(sgm->service_description);

			asprintf(&tmp,
				"INSERT INTO npc_servicegroup_membership ("
					"serviceid,"
					"groupid"
				") VALUES ((SELECT id from npc_service where service_description=%s and hostid = (SELECT id from npc_host where name=%s)),(SELECT id from npc_servicegroup where name=%s))",
					NULL_STRING(serviceName),
					NULL_STRING(hostName),
					NULL_STRING(groupName));
			// err handling
			buffer(&query_buf, tmp, ++query_buf_count);
			free(tmp);
			
			if(serviceName) free(serviceName);
		}

		if(groupName) free(groupName);

		groupcount++;
	}

	sprintf(log,"nagios-db: loaded %d servicegroups", groupcount);
	write_to_logs_and_console(log,NSLOG_INFO_MESSAGE,TRUE);
	
	/* also insert various config parameters that we might want */
	asprintf(&tmp, 
		"INSERT INTO npc_config_params ("
			"opt,"
			"value"
		") VALUES " 
			"('accept_passive_host_checks',%d),"
			"('execute_host_checks',%d),"
			"('accept_passive_service_checks',%d),"
			"('execute_service_checks',%d),"
			"('enable_flap_detection',%d),"
			"('enable_notifications',%d),"
			"('enable_event_handlers',%d)",
		accept_passive_host_checks,
		execute_host_checks,
		accept_passive_service_checks,
		execute_service_checks,
		enable_flap_detection,
		enable_notifications,
		enable_event_handlers);
	// err handling
	buffer(&query_buf, tmp, ++query_buf_count);
	free(tmp);

	write_to_logs_and_console("nagios-db: configuration complete. Registering for status notifications...",NSLOG_INFO_MESSAGE,TRUE);

	neb_deregister_callback(NEBCALLBACK_TIMED_EVENT_DATA, loadconfig);
	neb_register_callback(NEBCALLBACK_HOST_STATUS_DATA, inserter_module_handle, 0, processStatus);
	neb_register_callback(NEBCALLBACK_SERVICE_STATUS_DATA, inserter_module_handle, 0, processStatus);
	neb_register_callback(NEBCALLBACK_HOST_CHECK_DATA, inserter_module_handle, 0, processCheck);
	neb_register_callback(NEBCALLBACK_SERVICE_CHECK_DATA, inserter_module_handle, 0, processCheck);
	
	return 0;
}



static int processCheck(int cmd, void * data) 
{

	char *tmp;
	char log[2048];

	switch(cmd) 
	{
		case NEBCALLBACK_HOST_CHECK_DATA:
		{
			nebstruct_host_check_data * host_check_data = (nebstruct_host_check_data *) data;

			if(host_check_data->type != NEBTYPE_HOSTCHECK_PROCESSED) return 0;

			char *safeName = querify(host_check_data->host_name);

			asprintf(&tmp,
				"update npc_host set has_been_checked = TRUE where name = %s",
				NULL_STRING(safeName));
				// err handling
				buffer(&query_buf, tmp, ++query_buf_count);
				free(tmp);

			if(safeName) free(safeName);
		}
		break;

		case NEBCALLBACK_SERVICE_CHECK_DATA:
		{
			nebstruct_service_check_data * service_check_data = (nebstruct_service_check_data *) data;

			if(service_check_data->type != NEBTYPE_SERVICECHECK_PROCESSED) return 0;

			char *safeName = querify(service_check_data->host_name);
			char *safeServiceDescription = querify(service_check_data->service_description);

			asprintf(&tmp,
				"update npc_service set has_been_checked = TRUE where hostid = (SELECT id from npc_host where name=%s) and service_description = %s",
				NULL_STRING(safeName),
				NULL_STRING(safeServiceDescription));
				// err handling
				buffer(&query_buf, tmp, ++query_buf_count);
				free(tmp);

			if(safeName) free(safeName);
			if(safeServiceDescription) free(safeServiceDescription);
		}
    	break;

		default:
			sprintf(log,"nagios-db: processCheck ignoring unknown NEB command %i", cmd);
			write_to_logs_and_console(log,NSLOG_INFO_MESSAGE,TRUE);
			return 0;
		break;
	}

	return 0;
}



static int processStatus(int cmd, void * data) {

	service 	*tmp_service;
	host 		*tmp_host;
	time_t now;
	char *tmp;
	char log[2048];

	/* only at this point can we unregister our recurring event from the list */
	if (delayed_unload == TRUE){
			nebmodule_deinit(0,0);
			return -1;
	}
	
	switch(cmd) 
	{
		case NEBCALLBACK_HOST_STATUS_DATA:
		{
			nebstruct_host_status_data * host_status_data = (nebstruct_host_status_data *) data;
			tmp_host = (host *) host_status_data->object_ptr;
			char *safeName = querify(tmp_host->name);
			char *safeEventHandler = querify(tmp_host->event_handler);
			char *safePluginOutput = querify(tmp_host->plugin_output);
			char *safeHostCheckCommand = querify(tmp_host->host_check_command);
			char *safePerfData = querify(tmp_host->perf_data);
			
			/* host data we can safely bulk and buffer */
			
			/* only need the header once */
			if (bulk_buf[BULK_HOST] == NULL) {
				asprintf(&bulk_buf[BULK_HOST],
					"INSERT INTO npc_host (" 
						"name,"
						"active_checks_enabled,"
						"passive_checks_enabled,"
						"event_handler_enabled,"
						"flap_detection_enabled,"
						"notifications_enabled,"
						"failure_prediction_enabled,"
						"process_performance_data,"
						"obsess_over_host,"
						"event_handler,"
						"modified_attributes,"
						"has_been_checked,"
						"should_be_scheduled,"
						"configured"
					") VALUES ");
			}
			asprintf(&tmp, "(%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%lu,%s,%s,TRUE),",
				NULL_STRING(safeName),
				BOOLIFY_STRING(tmp_host->checks_enabled),
				BOOLIFY_STRING(tmp_host->accept_passive_host_checks),
				BOOLIFY_STRING(tmp_host->event_handler_enabled),
				BOOLIFY_STRING(tmp_host->flap_detection_enabled),
				BOOLIFY_STRING(tmp_host->notifications_enabled),
				BOOLIFY_STRING(tmp_host->failure_prediction_enabled),
				BOOLIFY_STRING(tmp_host->process_performance_data),
				BOOLIFY_STRING(tmp_host->obsess_over_host),
				NULL_STRING(safeEventHandler),
				tmp_host->modified_attributes,
				BOOLIFY_STRING(tmp_host->has_been_checked),
				BOOLIFY_STRING(tmp_host->should_be_scheduled));
			if (append(&bulk_buf[BULK_HOST], tmp, FALSE))
				err_out(ERR_QUERY_BUILD);
			free(tmp);
			
			/* check for empty check_commands */
			if (tmp_host->host_check_command) {
				/* header only once */
				if (bulk_buf[BULK_HOST_CHECK] == NULL) {
					asprintf(&bulk_buf[BULK_HOST_CHECK],
						"INSERT INTO npc_host_check ("
							"hostid,"
							"check_command,"
							"check_execute_time,"
							"check_latency,"
							"last_check,"
							"next_check,"
							"max_attempts,"
							"last_notification,"
							"next_notification,"
							"no_more_notifications,"
							"check_type,"
							"performance_data"
						") VALUES ");
				}	
				asprintf(&tmp, "((SELECT id from npc_host where name=%s), %s,%f,%f,%d,%d,%d,%d,%d,%s,%d,%s),",
					NULL_STRING(safeName),
					NULL_STRING(safeHostCheckCommand),
					tmp_host->execution_time,
					tmp_host->latency,
					(int) tmp_host->last_check,
					(int) tmp_host->next_check,
					tmp_host->max_attempts,
					(int) tmp_host->last_host_notification,
					(int) tmp_host->next_host_notification,
					BOOLIFY_STRING(tmp_host->no_more_notifications),
					tmp_host->check_type,
					NULL_STRING(safePerfData));
				if (append(&bulk_buf[BULK_HOST_CHECK], tmp, FALSE))
					err_out(ERR_QUERY_BUILD);
				free(tmp);
			} else {
				/* insert dummy values */
				if (bulk_buf[BULK_HOST_CHECK] == NULL) {
					asprintf(&bulk_buf[BULK_HOST_CHECK],
						"INSERT INTO npc_host_check ("
							"hostid,"
							"check_command,"
							"check_execute_time,"
							"check_latency,"
							"last_check,"
							"next_check,"
							"max_attempts,"
							"last_notification,"
							"next_notification,"
							"no_more_notifications,"
							"check_type,"
							"performance_data"
						") VALUES ");
				}	
				asprintf(&tmp, "((SELECT id from npc_host where name=%s), '',0,0,0,0,0,0,0,TRUE,0,NULL),",
					NULL_STRING(safeName));
				if (append(&bulk_buf[BULK_HOST_CHECK], tmp, FALSE))
					err_out(ERR_QUERY_BUILD);
				free(tmp);
			}
				
			/* now our various status tables that we can't bulk update, just buffer for
			 * for execution after the above updates run */
			
			/* saves us some subquerys */
			asprintf(&tmp,
				"set @hostid=(SELECT id from npc_host where name=%s)",
					NULL_STRING(safeName));
			// err handling
			buffer(&query_buf, tmp, ++query_buf_count);
			free(tmp);

			/* record flapping hosts */	
			if (tmp_host->is_flapping) {
				asprintf(&tmp,
					"INSERT IGNORE INTO npc_flapping_host ("
						"id"
					") VALUES (@hostid)");
			} else {
				asprintf(&tmp,
					"DELETE from npc_flapping_host where id=@hostid");
			}
			// err handling
			buffer(&query_buf, tmp, ++query_buf_count);
			free(tmp);

			/* host history updates. use some tricks to buffer this data properly, as 
			 * these statements depend on the large bulked inserts above and have to be
			 * executed after */
			
			time(&now);

			/* a unique service period is defined as hostid+started+soft_state+hard_state
			 * so if this is a new period from the previous one, use a new timestamp to 
			 * passthrough the unique key check, else use the old timestamp so we can 
			 * trigger an update on the duplicate key. not pretty but it works */
			asprintf(&tmp,
				"set @time="
				"if(ifnull((select hostid FROM npc_host_history WHERE hostid = @hostid and lasted is null and (soft_state != %d or hard_state != %d) order by started desc limit 1),-1)>0,%lu,ifnull((select started from npc_host_history where hostid = @hostid order by started desc limit 1),%lu))",
					tmp_host->current_state,
					tmp_host->last_hard_state,
					now, now);
			// error handling
			buffer(&query_buf, tmp, ++query_buf_count);
			free(tmp);

			/* starts new period or updates current_attempt depending on @time above */
			asprintf(&tmp,
				"INSERT INTO npc_host_history ("
					"hostid,"
					"started,"
					"lasted,"
					"soft_state,"
					"hard_state,"
					"current_attempt,"
					"plugin_output"
				") VALUES (@hostid,@time,null,%d,%d,%d,%s)"
				" ON DUPLICATE KEY UPDATE "
				"current_attempt=VALUES(current_attempt),"
				"plugin_output=VALUES(plugin_output)",
					tmp_host->current_state,
					tmp_host->last_hard_state,
					tmp_host->current_attempt,
					NULL_STRING(safePluginOutput));
			// error handling
			buffer(&query_buf, tmp, ++query_buf_count);
			free(tmp);

			asprintf(&tmp,
				"UPDATE npc_host_history set "
					"lasted=%lu - started where started != @time and hostid=@hostid and lasted is null",
					now);
			// error handling
			buffer(&query_buf, tmp, ++query_buf_count);
			free(tmp);
			
			if(safeName) free(safeName);
			if(safeEventHandler) free(safeEventHandler);
			if(safePluginOutput) free(safePluginOutput);
			if(safeHostCheckCommand) free(safeHostCheckCommand);
			if(safePerfData) free(safePerfData);
		}
		break;

		case NEBCALLBACK_SERVICE_STATUS_DATA:
		{
			nebstruct_service_status_data * service_status_data = (nebstruct_service_status_data *) data;
			tmp_service = (service *) service_status_data->object_ptr;

			char *safeName = querify(tmp_service->host_name);
			char *safeServiceDescription = querify(tmp_service->description);
			char *safeEventHandler = querify(tmp_service->event_handler);
			char *safePluginOutput = querify(tmp_service->plugin_output);
			char *safeServiceCheckCommand = querify(tmp_service->service_check_command);
			char *safePerfData = querify(tmp_service->perf_data);

			
			if (bulk_buf[BULK_SERVICE] == NULL) {
				asprintf(&bulk_buf[BULK_SERVICE],
					"INSERT INTO npc_service (" 
						"hostid,"
						"service_description,"
						"active_checks_enabled,"
						"passive_checks_enabled,"
						"event_handler_enabled,"
						"flap_detection_enabled,"
						"notifications_enabled,"
						"failure_prediction_enabled,"
						"process_performance_data,"
						"obsess_over_service,"
						"event_handler,"
						"modified_attributes,"
						"has_been_checked,"
						"should_be_scheduled,"
						"configured"
					") VALUES ");
			}	
			asprintf(&tmp, "((SELECT id from npc_host where name=%s),%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%lu,%s,%s,TRUE),",
				NULL_STRING(safeName),
				NULL_STRING(safeServiceDescription),
				BOOLIFY_STRING(tmp_service->checks_enabled),
				BOOLIFY_STRING(tmp_service->accept_passive_service_checks),
				BOOLIFY_STRING(tmp_service->event_handler_enabled),
				BOOLIFY_STRING(tmp_service->flap_detection_enabled),
				BOOLIFY_STRING(tmp_service->notifications_enabled),
				BOOLIFY_STRING(tmp_service->failure_prediction_enabled),
				BOOLIFY_STRING(tmp_service->process_performance_data),
				BOOLIFY_STRING(tmp_service->obsess_over_service),
				NULL_STRING(safeEventHandler),
				tmp_service->modified_attributes,
				BOOLIFY_STRING(tmp_service->has_been_checked),
				BOOLIFY_STRING(tmp_service->should_be_scheduled));
			if (append(&bulk_buf[BULK_SERVICE], tmp, FALSE))
					err_out(ERR_QUERY_BUILD);
			free(tmp);

			/* load up service id */
			
			if (bulk_buf[BULK_SERVICE_CHECK] == NULL) {
				asprintf(&bulk_buf[BULK_SERVICE_CHECK],
					"INSERT INTO npc_service_check (" 
						"serviceid,"
						"check_command,"
						"check_execute_time,"
						"check_latency,"
						"last_check,"
						"next_check,"
						"max_attempts,"
						"last_notification,"
						"next_notification,"
						"no_more_notifications,"
						"check_type,"
						"performance_data"
					") VALUES ");
			}
			asprintf(&tmp, "((SELECT id from npc_service where service_description=%s and hostid=(SELECT id from npc_host where name=%s)),"
					"%s,%f,%f,%d,%d,%d,%d,%d,%s,%d,%s"
					"),",
				NULL_STRING(safeServiceDescription),
				NULL_STRING(safeName),
				NULL_STRING(safeServiceCheckCommand),
				tmp_service->execution_time,
				tmp_service->latency,
				(int) tmp_service->last_check,
				(int) tmp_service->next_check,
				tmp_service->max_attempts,
				(int) tmp_service->last_notification,
				(int) tmp_service->next_notification,
				BOOLIFY_STRING(tmp_service->no_more_notifications),
				tmp_service->check_type,
				NULL_STRING(safePerfData));
			if (append(&bulk_buf[BULK_SERVICE_CHECK], tmp, FALSE))
					err_out(ERR_QUERY_BUILD);
			free(tmp);

			/* now our various status tables that we can't bulk update, just buffer for
			 * for execution after the above updates run */
			
			/* saves us some subquerys */
			asprintf(&tmp,
				"set @serviceid=(SELECT id from npc_service where service_description=%s and hostid=(SELECT id from npc_host where name=%s))",
					NULL_STRING(safeServiceDescription),
					NULL_STRING(safeName));
			// err handling
			buffer(&query_buf, tmp, ++query_buf_count);
			free(tmp);

			/* record flapping services */	
			if (tmp_service->is_flapping) {
				asprintf(&tmp,
					"INSERT IGNORE INTO npc_flapping_service ("
						"id"
					") VALUES (@serviceid)");
			} else {
				asprintf(&tmp,
					"DELETE from npc_flapping_service where "
						"id=@serviceid");
			}
			// err handling
			buffer(&query_buf, tmp, ++query_buf_count);
			free(tmp);

			/* service history updates. use some tricks to buffer this data properly, as 
			 * these statements depend on the large bulked inserts above and have to be
			 * executed after */
			
			time(&now);

			/* a unique service period is defined as serviceid+started+soft_state+hard_state
			 * so if this is a new period from the previous one, use a new timestamp to 
			 * passthrough the unique key check, else use the old timestamp so we can 
			 * trigger an update on the duplicate key. not pretty but it works */
			asprintf(&tmp,
				"set @time="
				"if(ifnull((select serviceid FROM npc_service_history WHERE serviceid = @serviceid and lasted is null and (soft_state != %d or hard_state != %d) order by started desc limit 1),-1)>0,%lu,ifnull((select started from npc_service_history where serviceid = @serviceid order by started desc limit 1),%lu))",
					tmp_service->current_state,
					tmp_service->last_hard_state,
					now, now);
			// error handling
			buffer(&query_buf, tmp, ++query_buf_count);
			free(tmp);

			/* starts new period or updates current_attempt depending on @time above */
			asprintf(&tmp,
				"INSERT INTO npc_service_history ("
					"serviceid,"
					"started,"
					"lasted,"
					"soft_state,"
					"hard_state,"
					"current_attempt,"
					"plugin_output"
				") VALUES (@serviceid,@time,null,%d,%d,%d,%s)"
				" ON DUPLICATE KEY UPDATE "
				"current_attempt=VALUES(current_attempt),"
				"plugin_output=VALUES(plugin_output)",
					tmp_service->current_state,
					tmp_service->last_hard_state,
					tmp_service->current_attempt,
					NULL_STRING(safePluginOutput));
			// error handling
			buffer(&query_buf, tmp, ++query_buf_count);
			free(tmp);

			asprintf(&tmp,
				"UPDATE npc_service_history set "
					"lasted=%lu - started where started != @time and serviceid=@serviceid and lasted is null",
					now);
			// error handling
			buffer(&query_buf, tmp, ++query_buf_count);
			free(tmp);

			if(safeName) free(safeName);
			if(safeServiceDescription) free(safeServiceDescription);
			if(safeEventHandler) free(safeEventHandler);
			if(safePluginOutput) free(safePluginOutput);
			if(safeServiceCheckCommand) free(safeServiceCheckCommand);
			if(safePerfData) free(safePerfData);
		}
    	break;

		default:
			sprintf(log,"nagios-db: processCheck ignoring unknown NEB command %i", cmd);
			write_to_logs_and_console(log,NSLOG_INFO_MESSAGE,TRUE);
			return 0;
		break;
	}

	return 0;
}

/* helper function to make a string ready for inclusion into a sql query */
char *
querify(char * src)
{
	char * ret = 0;
	if(src)
	{
		int size = 2*strlen(src) + 3; // we want to surround strings with single quotes
		int written = 0;
		if(!(ret = malloc(size)))
		{
			// error handling 
			return 0;
		}
		ret[0] = '\'';
		written = mysql_real_escape_string(&mysql, ret+1, src, strlen(src));
		ret[written + 1] = '\'';
		ret[written + 2] = 0;
	}
	return ret;
}

/* resize query buffer and add a new entry */
int buffer(char ***dest, char *src, int i)
{
	char **tmp;

	// is this necessary, can we really not realloc a non malloc'd block?
	if (i - 1 == 0) {
		tmp = (char **)malloc(sizeof (char *) * (i));
	} else {
		tmp = (char **)realloc(*dest, sizeof (char *) * (i));
	}
	// err handling here!
	
	tmp[i - 1] = strdup(src);
	*dest = tmp;

	return 0;
}

/* resize and append string */
int append(char **dest, char *src, int remove_comma)
{
	char *tmp;

	tmp = (char *)realloc(*dest, sizeof (char *) * (strlen(*dest) + strlen(src) +  1));

	if (tmp==NULL) {
		write_to_logs_and_console("Could not allocate memory",NSLOG_INFO_MESSAGE,TRUE);
		return -1;
	} else {
		if (remove_comma)
			tmp[strlen(tmp)-1] = ' ';
		strcat(tmp, src);
		*dest = tmp;
	}	
	
	return 0;
}

void free_buffers()
{

	int i;
	for (i = 0; i < bulk_buf_count; i++) {	
		free(bulk_buf[i]);
		bulk_buf[i]=NULL;
	}

	if (query_buf_count) {
		while (--query_buf_count) {
			free(query_buf[query_buf_count - 1]);
	}
		
	free(query_buf);
	query_buf = NULL;
	}
}

void err_out (char *reason) 
{
	write_to_logs_and_console(reason,NSLOG_INFO_MESSAGE,TRUE);

	free_buffers();
	
	delayed_unload=TRUE;
}

static int sql_dump () 
{
	int i;
	char *tmp;

	/* finish up bulk bufferd inserts */
	if (bulk_buf[BULK_HOST]) {
		if (append(&bulk_buf[BULK_HOST],
				" ON DUPLICATE KEY UPDATE " 
				"active_checks_enabled=VALUES(active_checks_enabled),"
				"passive_checks_enabled=VALUES(passive_checks_enabled),"
				"event_handler_enabled=VALUES(event_handler_enabled),"
				"flap_detection_enabled=VALUES(flap_detection_enabled),"
				"notifications_enabled=VALUES(notifications_enabled),"
				"failure_prediction_enabled=VALUES(failure_prediction_enabled),"
				"process_performance_data=VALUES(process_performance_data),"
				"obsess_over_host=VALUES(obsess_over_host),"
				"event_handler=VALUES(event_handler),"
				"modified_attributes=VALUES(modified_attributes),"
				"has_been_checked=VALUES(has_been_checked),"
				"should_be_scheduled=VALUES(should_be_scheduled),"
				"configured=TRUE",
			TRUE))
			err_out(ERR_QUERY_BUILD);
	}
	if (bulk_buf[BULK_HOST_CHECK]) {
		if (append(&bulk_buf[BULK_HOST_CHECK],
				" ON DUPLICATE KEY UPDATE "
				"check_execute_time=VALUES(check_execute_time),"
				"check_latency=VALUES(check_latency),"
				"last_check=VALUES(last_check),"
				"next_check=VALUES(next_check),"
				"max_attempts=VALUES(max_attempts),"
				"last_notification=VALUES(last_notification),"
				"next_notification=VALUES(next_notification),"
				"no_more_notifications=VALUES(no_more_notifications),"
				"check_type=VALUES(check_type),"
				"performance_data=VALUES(performance_data)",
			TRUE))
			err_out(ERR_QUERY_BUILD);
	}
	if (bulk_buf[BULK_SERVICE]) {
		if (append(&bulk_buf[BULK_SERVICE],
				" ON DUPLICATE KEY UPDATE "
				"active_checks_enabled=VALUES(active_checks_enabled),"
				"passive_checks_enabled=VALUES(passive_checks_enabled),"
				"event_handler_enabled=VALUES(event_handler_enabled),"
				"flap_detection_enabled=VALUES(flap_detection_enabled),"
				"notifications_enabled=VALUES(notifications_enabled),"
				"failure_prediction_enabled=VALUES(failure_prediction_enabled),"
				"process_performance_data=VALUES(process_performance_data),"
				"obsess_over_service=VALUES(obsess_over_service),"
				"event_handler=VALUES(event_handler),"
				"modified_attributes=VALUES(modified_attributes),"
				"has_been_checked=VALUES(has_been_checked),"
				"should_be_scheduled=VALUES(should_be_scheduled),"
				"configured=TRUE",
			TRUE))
			err_out(ERR_QUERY_BUILD);
	}
	if (bulk_buf[BULK_SERVICE_CHECK]) {
		if (append(&bulk_buf[BULK_SERVICE_CHECK],
				" ON DUPLICATE KEY UPDATE " 
				"check_execute_time=VALUES(check_execute_time),"
				"check_latency=VALUES(check_latency),"
				"last_check=VALUES(last_check),"
				"next_check=VALUES(next_check),"
				"max_attempts=VALUES(max_attempts),"
				"last_notification=VALUES(next_notification),"
				"next_notification=VALUES(next_notification),"
				"no_more_notifications=VALUES(no_more_notifications),"
				"check_type=VALUES(check_type),"
				"performance_data=VALUES(performance_data)",
			TRUE))
			err_out(ERR_QUERY_BUILD);
	}
	for (i = 0; i < bulk_buf_count; i++) {	
		if (bulk_buf[i]) {
			// todo: debug here
			//printf("bulk_buf_count %i : incoming query is %i bytes long\n",i,strlen(bulk_buf[i]) * sizeof (char *));
			// todo: if length of query bigger than max_packet, set it higher for this session, ie @@session.max_allowed_packet=16777216;
			if (mysql_real_query(&mysql, bulk_buf[i], strlen(bulk_buf[i]))) {
	 			printf("\n%s %i\n%s",mysql_error(&mysql),mysql_errno(&mysql),bulk_buf[i]);
				exit(1);
	  		}
		}

	}

	// todo: debug here
	//printf("query_buf_count %i\n", query_buf_count);
	for (i = 0; i < query_buf_count; i++) {
		if (mysql_real_query(&mysql, query_buf[i], strlen(query_buf[i]))) {
	 		asprintf(&tmp, "%s %i\n%s",mysql_error(&mysql),mysql_errno(&mysql),query_buf[i]);
			err_out(tmp);
			free(tmp);
	  		return -1;
		}
	}
	free_buffers();
	return 0;
}
